package MCS::CGI::Conf;

use strict;
use vars qw( $VERSION @ISA $global_cfg_env $default_cfg_env $adm_group $debug $verbose_insecure);

$debug=0;
$verbose_insecure=1;

##
## MCS::CGI::Conf -- Konfigurationsdateien fuer Standard-CGIs bei MCS
##
## $Id: Conf.pm,v 1.25 2007/08/30 09:48:45 sebi Exp $
## Copyright (C) 2000 MCS Moorbek Computer Systeme GmbH
## Author: Geoff Simmons <gs@cityline.net>
##         Nico Tillmann <nico@mcs-cityline.net>
##
## $Log: Conf.pm,v $
## Revision 1.25  2007/08/30 09:48:45  sebi
## * fixed: debug und inscure file permission reporting sind nun getrennt steuerbar.
##

#use IniConf;
use Config::IniFiles;
use Carp;
use Fcntl qw( :flock);

require Exporter;

@ISA = qw(Exporter);
$VERSION = sprintf "%d.%02d", (q$Revision: 1.25 $ =~ /\d+/g);

=head1 NAME

MCS::CGI::Conf - Konfigurationsdateien fuer Standard-CGIs bei MCS

=head1 SYNOPSIS

  use MCS::CGI::Conf;

  my $cfg  = new MCS::CGI::Conf;

  # Ermittelt den Wert von Parameter 'bar' im Abschnitt 'foo'
  my $foo_bar = $cfg->val('foo', 'bar');

=head1 DESCRIPTION

Das Modul B<MCS::CGI::Conf> stellt Funktionen zur VerfE<uuml>gung, die
die Arbeit mit Konfigurationsdateien fE<uuml>r die
Standard-CGI-Programme von MCS unterstutzen. Die Funktionen
E<uuml>berprE<uuml>fen die Sicherheit von Konfigurationsdateien, lesen
ihre Inhalte ein und geben ihre Parameterwerte zurE<uuml>ck.

=head2 Format

Das Format der Konfig-Dateien E<auml>hnelt dem Format von C<.ini>-Dateien
in der Windows-Welt:

  # Zuerst die Werte von Abschnitt 1
  [abschnitt1]
  var1 = val1
  var2 = val2  ;; Ziemlich naheliegend

  ;; Semikolon und '#' leiten Kommentare ein
  [abschnitt2]
  var1 = val1  # Keine Wiederholung, weil wir uns
  var2 = val2  # in Abschnitt 2 befinden.

Die Dateien werden in verschiedene Abschnitte aufgeteilt, die mit Symbolen
in eckigen Klammern eingeleitet werden
(z.B. C<[mail]>). Parameterzuweisungen werden mit der Syntax C<variable =
wert> geschrieben. Kommentare werden mit dem Semikolon oder dem '#'
eingeleitet; vgl. L<IniConf>.

=head2 Funktionen

=over

=item B<new>

  my $cfg = new MCS::CGI::Conf;

Der Konstruktor veranlasst, daE<szlig> drei Konfigurationsdateien gelesen
werden, und zwar die globale, lokale und Default-Konfig-Dateien. Ihre Pfade
werden E<uuml>ber Environment-Variablen festgelegt;
s. L</Konfigurationsdateien>.

Beim Einlesen werden die Dateien zuerst nach Sicherheit geprE<uuml>ft:
S. dazu L</Sicherheit>. Kann ihre Sicherheit nicht geprE<uuml>ft
werden (weil der Befehl C<stat> nicht ausgefE<uuml>hrt werden konnte,
s. L<perlfunc/stat>), oder wenn die Sicherheitskriterien nicht
erfE<uuml>llt werden kE<ouml>nnen, dann endet das Programm mit einem
Aufruf von C<Carp::croak> (s. L<Carp>). Allerdings kann mit C<eval>
verhindert werden, daE<szlig> das Programm gestoppt wird
(s. L<perlfunc/eval>):

  eval { $cfg = new MCS::CGI::Conf };
  warn "Could not read configuration: $@\n"
    if $@;

=item B<val> I<ABSCHNITT> I<VARIABLE>

  my $mailprog = $cfg->val('mail', 'sendmail');

Die Funktion C<val> gibt den Wert zurE<uuml>ck, der der I<VARIABLE> im
genannten I<ABSCHNITT> zugewiesen wurde.

=item B<LocalSections>

Gibt ein Array mit den Sectionsnamen der lokalen Konfigurationsdatei zurE<uuml>ck.


=cut


sub LocalSections{
    my ($self) = @_;
    return $self->{local}->Sections;
}

=item B<LocalParameters> ($sectionname)

Gibt ein Array mit den Parameter einer spezifischen localen Section zurueck.

=cut

sub LocalParameters {
  my ($self, $section) = @_;
  return
      defined $self->{local}->Parameters($section) ?
	  $self->{local}->Parameters($section) : ();
}


=item B<is_safe> ($configfile)

=over

=item * falls  HTTP_MCS_NO_SUEXEC eq true

wir lediglich geprueft, ob die Datei nicht schreibbar fuer other ist.
Da Datei fuer User des Webserver
e.g. httpd lesbar sein muss, ist in der Konfiguration des Webservers
dafuer Sorge zu tragen, dass die Konfigurationsdatein
(typischer weise .conf oder .ini)
nicht ausgeliefert werden.

=item * andernfalls

prueft ob die Konfigurationsdatei die unter dem Punkt B<Sicherheit> erwaehnten Kriterien erfuellt.

=back

=cut

sub is_safe {
  my $file = shift;
  my @stat = stat($file);
  unless(@stat) {
      print STDERR "DEBUG_HINT:", __FILE__," ", __LINE__,"\n", if ($debug);
      print STDERR "stat($file) failed - probably file doesn't exist\n", if ($verbose_insecure);
      return undef;
  }

  if( $ENV{HTTP_MCS_NO_SUEXEC} && $ENV{HTTP_MCS_NO_SUEXEC} eq "true") {
      # Darf nicht fuer 'other' schreibbar sein
      if ($stat[2] & 002) {
	print STDERR __FILE__,__LINE__,"\n$file ist schreibbar fuer other\n", if ($verbose_insecure);
	  return 0;
      }
      print STDERR __FILE__,__LINE__,"\n$file ist nicht schreibbar fuer other und HTTP_MCS_NO_SUEXEC ist true\nis_safe gibt 1 zurueck\n" if($debug);
      return 1;
  }



  # Owner muss root sein oder mit EUID uebereinstimmen
  if( $stat[4] != $> and $stat[4] != 0) {
      print STDERR __FILE__,__LINE__,"\n$stat[4] != $> and $stat[4] != 0\n" if($verbose_insecure);
	  return 0;
  }

  # Darf nicht fuer 'other' schreib- oder lesbar sein
  if(  $stat[2] & 006) {
    print STDERR __FILE__,__LINE__,"\n$stat[2] & 006\n", if ($verbose_insecure);
    return 0;
  }

  # Wenn Gruppe nicht 'wwwadm' ist, darf sie nicht fuer die Gruppe
  # schreib- oder lesbar sein.
  if ($stat[2] & 060) {
    my %adm_groups;
    for (split /:/, $ENV{$adm_group}) {
      $adm_groups{$_}++;
    }
    unless ( $adm_groups{getgrgid($stat[5])} ) {
      print STDERR __FILE__,__LINE__,"\n$adm_groups{getgrgid($stat[5])}\n", if ($verbose_insecure);
       return 0;
    }
  }
#   return 0
#     if getgrgid($stat[5]) ne $ENV{$adm_group} and $stat[2] & 060;
  print STDERR __FILE__,__LINE__,"\nis_safe gibt 1 zurueck\n" if($debug);
  return 1;
}


=back

=head2 Konfigurationsdateien

Eine Unterscheidung zwischen I<globalen>, I<lokalen> und
I<Default->Konfig-Dateien wir berE<uuml>cksichtigt:

=over

=item Globale Konfig-Datei

Der Pfad zur globalen Konfig-Datei wird in der Environment-Variable
C<HTTP_MCS_GLOBAL_CFG> angegeben. Sie definiert Werte, die fE<uuml>r alle
virtuelle Hosts an einer physikalischen Site gelten sollen, und die
mE<ouml>glicherweise empfindlich fE<uuml>r die Serversicherheit
sind. Diese Datei soll nicht von Kunden, sondern nur von
Administratoren editierbar sein. Sie soll nicht unter dem
Dokumentenpfad von irgendeinem Web-Server liegen.

In der globalen Konfig-Datei kE<ouml>nnte z.B. den Pfad zum
Sendmail-Programm festgelegt werden. Ein mE<ouml>glicher Ort fE<uuml>r
die Datei kE<ouml>nnte F</var/httpd/support/std_cgi_global.cfg> sein.

=item Default Konfig-Datei

Der Pfad zur Default-Konfig-Datei wird in der Environment-Variable
C<HTTP_MCS_DEFAULT_CFG> angegeben. Sie ist der globalen Konfig-Datei
E<auml>hnlich, da sie Werte definiert, die fE<uuml>r alle virtuelle
Hosts an einer Site gelten, und nicht von Kunden, sondern nur von
Administratoren editierbar ist. Aber die darin definierten Werte
kE<ouml>nnen von lokalen Konfig-Dateien E<uuml>berschrieben werden;
diese Werte sind also systemweite Defaults.

In der Default-Konfig-Datei kE<ouml>nnte z.B. eine Email-Adresse fE<uuml>r
die Benachrichtigung E<uuml>ber Server-Fehler eingetragen werden. Ein
mE<ouml>glicher Ort fE<uuml>r die Datei kE<ouml>nnte
F</var/httpd/support/std_cgi_default.cfg> sein.

=item Lokale Konfig-Dateien

Die lokalen Konfig-Dateien werden in der Environment-Variable
C<PATH_TRANSLATED> angegeben. Sie definieren Werte, die fE<uuml>r den
jeweiligen virtuellen Server gelten. Sie mE<uuml>ssen im
Dokumentenpfad des jeweiligen Servers liegen, und kE<ouml>nnen vom
Kunden editierbar sein.

In der lokalen Konfig-Datei kE<ouml>nnte z.B. die HTML-Seiten
festgelegt werden, zu denen der Server bei HTTP Status 500 oder 404
(Internal Error bzw. File Not Found) verzweigen soll.

=back

=head2 Sicherheit

Folgende Kriterien mE<uuml>ssen gelten, sonst werden die Dateien als
unsicher angesehen:

=over

=item 1

Der Owner der Datei muss C<root> sein, oder muss mit der effektiven
UserID des laufenden Programms E<uuml>bereinstimmen.

=item 2

Die Datei darf nicht fE<uuml> 'other' schreib- oder lesbar sein.

=item 3

Wenn die Gruppe der Datei nicht eine der administrativen Gruppen von MCS
ist, dann darf sie auch nicht fE<uuml>r die Gruppe schreib- oder lesbar
sein.

Die administrativen Gruppen werden in der Environment-Variable
C<HTTP_MCS_ADMIN_GRP> definiert. Gibt es mehr als es eine Gruppe, so
kE<ouml>nnen sie mit 'C<:>' getrennt werden, z.B. C<wwwadm:suexec>.

=back

Besteht die Datei alle SicherheitsprE<uuml>fungen, dann wird sein Name
vor dem Einlesen 'ge-untaint-et' (s. L<perlsec>). Das bedeutet,
daE<szlig> die eingelesenen Werte auch bei Taint-Checking als sicher
gelten.

=head2 Environment

Auf der globalen Ebene der Server-Konfiguration (d.h. in der Konfiguration,
die fE<uuml>r alle virtuellen Hosts gelten) sollten die
Environment-Variablen C<HTTP_MCS_GLOBAL_CFG>, C<HTTP_MCS_DEFAULT_CFG> und
C<HTTP_MCS_ADMIN_GRP> gesetzt werden, z.B. so:

  ## In httpd.conf

  # Die globale Konfig-Datei, die fuer alle virtuellen Hosts gilt
  SetEnv HTTP_MCS_GLOBAL_CFG /var/httpd/support/std_cgi_global.cfg

  # Die globale Default-Konfig-Datei, die fuer alle virtuellen Hosts gilt
  SetEnv HTTP_MCS_DEFAULT_CFG /var/httpd/support/std_cgi_default.cfg

  # Die administrativen Gruppen, die Konfig-Dateien editieren darf
  SetEnv HTTP_MCS_ADMIN_GRP wwwadm:suexec

  # optional die Angabe ob HTTPD mit SUEXEC laueft, falls die Variable
  # fehlt wird HTTP_MCS_NO_SUEXEC false angenommen
  SetEnv HTTP_MCS_NO_SUEXEC true|false

Zu der Konfig-Anweisung C<SetEnv> von Apache siehe:
http://www.apache.de/docs/mod/mod_env.html#setenv

Um eine lokale Konfig-Datei zu benutzen, muE<szlig> ihr Pfad mit der
Environemnt-Variable C<PATH_TRANSLATED> festgelegt werden. Das
erreicht man mit dem Aufruf des CGI-Programms:

  http://www.domain.de/cgi-bin-std/prog/conf/prog.cfg

Ist in diesem Beispiel F<prog> das CGI-Programm, und ist
F</var/httpd/servers/www.domain.de/htdocs> der Dokumentenpfad des
virtuellen Hosts, dann hat C<PATH_TRANSLATED> den Wert
F</var/httpd/servers/www.domain.de/htdocs/conf/prog.cfg>.

=cut

# Mit dem ref-to-glob Trick fungieren $global_cfg_env,
# $default_cfg_env und $adm_group als Konstanten; s. dazu den Kamel.

*global_cfg_env = \'HTTP_MCS_GLOBAL_CFG';
*default_cfg_env = \'HTTP_MCS_DEFAULT_CFG';
*adm_group = \'HTTP_MCS_ADMIN_GRP';
*no_suexec_env = \'HTTP_MCS_NO_SUEXEC';


sub safety_it{
   my $file = shift;
   my @stat = stat($file);
   return unless @stat;

   if($file =~ /([^%&]*)/ ){
       $file = $1;
   }
   else{
       croak "Cannot check security of $file File contains invalid signs: %& : $!\n"
   }

   # Darf nicht fuer 'other' schreib- oder lesbar sein
   if( $stat[2] & 006 ) {

       #Datei auf o-rwx korregieren
       my $mode = ($stat[2] & 07770  );
       if( $mode =~ /(\d+)/) {
	   $mode = $1;
       }
       else{
	   $mode = 0640;
       }
       print STDERR __FILE__ ,": Dateirechte werden geaendert\nchmod $mode $file\n";
       return  chmod $mode, $file;

   }

}

sub read_config {
  my $cfg = shift
    or return Config::IniFiles->new(-file => '/dev/null');

  my $safe = is_safe($cfg);
  if( defined $safe && !$safe ){
      if( ($cfg =~ /\.ini$/) || ($cfg =~ /\.conf$/) ) {
          $safe = safety_it($cfg);
      }
  }
  my @file;
  my $has_dos_cr = 0;
  croak "Cannot check security of $cfg: $!\n"
    unless defined $safe;

  croak "$cfg is insecure\n"
    unless $safe;

  # Die Datei ist sicher, kann also ge-untaint-et werden
  $cfg =~ /(.*)/;
  $cfg = $1;

  # DOS Carriage Returns umwandeln
  unless( open FILE, "<$cfg" ) {
      print STDERR __FILE__, " at ",__LINE__, "Kann Datei $cfg nicht zum Lesen offnen\n", if ($debug);
  }
  else{
      flock(FILE, LOCK_SH );
      @file = <FILE>;
      flock(FILE, LOCK_UN);
      close FILE;

      foreach (@file) {
	  if ($_ =~ m/\r/g) {
	      s/\r//g;
	      $has_dos_cr = 1;
	  }
      }

      if($has_dos_cr) {
	  unless( open OUT,">$cfg" ) {
	      print STDERR __FILE__, " at ",__LINE__, "Kann Datei $cfg nicht zum Schreiben oeffnen\n", if ($debug);
	  }
	  else{
	      flock(OUT, LOCK_EX );
	      print OUT @file;
	      flock(OUT, LOCK_UN );
	      close OUT;
	  }
      }
  }
  my $iniconf = Config::IniFiles->new(-file => $cfg);

  if ( ! defined $iniconf ) {
    print STDERR "Fehler: IniConf->new returned undef.\n--- \@IniConf::errors:\n", if ($debug);
    print STDERR join("\n", @IniConf::errors), if ($debug);
    print STDERR "--- \@IniConf::errors end.\n", if ($debug);
  }

  return $iniconf;

}

sub new {
  my $class = shift;
  my $self = {};

  $self->{local}   = read_config($ENV{PATH_TRANSLATED});
  $self->{default} = read_config($ENV{$default_cfg_env});
  $self->{global}  = read_config($ENV{$global_cfg_env});

  unless( $self->{local} ) { print STDERR "read_config($ENV{PATH_TRANSLATED})\n schlug fehl; es wurde kein Objekt erzeugt\n", if ($debug);}
  unless( $self->{default} ) { print STDERR "read_config($ENV{$default_cfg_env})\n schlug fehl; es wurde kein Objekt erzeugt\n", if ($debug);}
  unless( $self->{global} ) { print STDERR "read_config($ENV{$global_cfg_env})\n schlug fehl; es wurde kein Objekt erzeugt\n", if ($debug);}

  return bless $self, $class;
}

sub val {
  my ($self, $section, $param) = @_;

  return
    defined $self->{global}->val($section, $param) ?
      $self->{global}->val($section, $param) :
	defined $self->{local}->val($section, $param) ?
	  $self->{local}->val($section, $param) :
	    $self->{default}->val($section, $param);
}

=pod

=head1 VERSION

$Id: Conf.pm,v 1.25 2007/08/30 09:48:45 sebi Exp $

=head1 AUTHOR

  Copyright (C) 2000 MCS Moorbek Computer Systeme GmbH
  Autor(en): Geoff Simmons <gs@mcs-cityline.net>
             Nico Tillmann <nico@mcs-cityline.net>
=head1 SEE ALSO

L<perl(1)>, L<perlsec>, L<Config::IniFiles>, L<perlfunc>, L<Carp>,
"Das Perl Kochbuch" Abs. 8.17,
http://www.apache.de/docs/mod/mod_env.html#setenv

=cut

1;

__END__
