package MCS::CGI::HTML::Template;

##
## Format an HTML template with string macros, conditionals and loops
##
## Copyright (C) 1999 Geoffrey Simmons <gs@cityline.net>
## This is free software. You may modify and/or redistribute it under
## the same terms as Perl itself.
##

# $Id: Template.pm,v 1.3 2007/06/21 13:28:19 sebi Exp $
# $Log: Template.pm,v $
# Revision 1.3  2007/06/21 13:28:19  sebi
# caller info on error msg added
#
# Revision 1.2  2002/07/04 13:09:52  nico
# HTML::Parser durch MCS::CGI::HTML::Parser und
# HTML::Filter durch MCS::CGI::HTML::Filter
# ersetzt.
# MCS::CGI::HTML::Template entsprechend angepasst
#
# Revision 1.1  2002/04/25 10:14:07  nico
# Umwandlung HTML::Template in MCS::CGI::HTML::Template
#
# Es muss nun noch eine Anpassung erfolgen, dass das Modul mit neuereren Versionen
# von HTML::Parser (bisher max 2.25) zusammenarbeiten kann.
#

# 
#
# Revision 1.6  1999/10/22 22:50:00  geoff
# Major rewrite so that nested IF's and WHILE's are allowed.
#
# Revision 1.5  1999/06/15 15:52:59  geoff
# Macros really are expanded within comments, as promised by the docs.
#
# Revision 1.4  1999/04/23 21:57:55  geoff
# Fixed a bug in generic() that was causing 'Uninitialized value' warnings
#
# Revision 1.3  1999/04/20 20:52:36  geoff
# - Edited the documentation
# - Now I really do have an inheritance-safe way to emit the package name.
# - Fixed a bug handling comments that are just comments.
#
#

use strict;
use vars qw($VERSION @ISA);
local $^W = 1; # Like switching on -w
               # local() restores the value of $^W outside the package

require MCS::CGI::HTML::Filter;
@ISA = qw(MCS::CGI::HTML::Filter);
$VERSION = sprintf "%d.%02d", (q$Revision: 1.3 $ =~ /\d+/g);

=head1 NAME

MCS::CGI::HTML::Template - Yet another template formatter for HTML sources

=head1 SYNOPSIS

  use MCS::CGI::HTML::Template;

  # Create an MCS::CGI::HTML::Template parser object and initialize some macro
  # values
  $p = new MCS::CGI::HTML::Template
  	DATE   => scalar(localtime),
  	MARKUP => 'STRONG';

  # Format a snippet of HTML and print the result 
  # This displays today's date with strong emphasis
  $p->parse('Today is <__MARKUP__>__DATE__</__MARKUP__>');

  # Add some more macro values
  $p->var(AUTHOR => 'Geoff', LANGUAGE => 'Perl');

  # Set up a conditional to return true if the value of REMOTE_USER is
  # 'larry'. The corresponding HTML will only display if it returns
  # true.
  sub user_larry { return $ENV{REMOTE_USER} eq 'larry' }
  $p->var(USER_LARRY => \&user_larry);

  # Set up a loop to show all environment variables, using an array of
  # anon hashes for the macro values.
  # See the DESCRIPTION for an alternative technique.
  @env = ();
  for (keys %ENV) { push @env, { VAR => $_, VAL => $ENV{$_} } }
  $p->var(ENV => \@env);

  # Format an HTML template and print the result
  $fh = new IO::File 'template.tmpl'
    or die "Cannot open 'template.tmpl': $!";
  $p->parse_file($fh);

  ##
  ## The template 'template.tmpl' might include the following:
  ##

  The author is <STRONG>__AUTHOR__</STRONG>,
  and the program was written in <EM>__LANGUAGE__</EM>.

  <!-- if USER_LARRY -->
  Hi Larry!
  <!-- end if -->

  <TABLE>
  <!-- while ENV -->
  <TR>
    <TH>__VAR__</TH>
    <TD>__VAL__</TD>
  </TR>
  <!-- end while -->
  </TABLE>

=head1 DESCRIPTION

The module B<MCS::CGI::HTML::Template> enforces a clean separation of
programming for the World Wide Web from screen layout using HTML. The
layout is confined to a I<template>, which contains HTML markup but
includes macros for data to be filled in by programs, as well as
instructions for portions of the page to displayed conditionally or
repeatedly. Thus development of the layout can be done solely by
working with the template, without necessarily getting the programs
involved.

=head2 HTML Template Syntax

The HTML template contains macros that are delimited by double
underscores:

    The date and time are <__EMPHASIS__>__DATE__</__EMPHASIS__>.
    <!-- Data for user __USER__ -->
    <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML __HTML_VERSION__//__LANG__">

Note that macros may appear anywhere -- not only in free text, but
also in start and end tags, comments, and the DTD document
declaration.

The template may also contain instructions for conditionals or loops
within HTML comments, as in:

    <!-- if MY_CONDITION -->
    The value is __VALUE__.
    <!-- end if -->

    <!-- while MY_LOOP -->
    <TR>
      <TH>__Key__</TH>
      <TD>__Value__</TH>
    </TR>
    <!-- end while -->

The identifiers in the C<if> and C<while> constructs associate that
portion of the template with Perl code that evaluates the conditional
or performs the loop iterations, as explained below. These comments
are not reproduced in the formatted HTML output.

The C<if> and C<while> constructs may be nested, like this:

  <!-- if MY_CONDITION -->
    <!-- while MY_LOOP -->
    <TR><TH>__Key__</TH><TD>__Value__</TD></TH></TR>
    <!-- end while -->
  <!-- end if -->

The identifiers (the macro names and the identifiers associated with
conditionals and loops) consist of alphanumeric and underscore
characters (thus they match the expression C<\w+>), and are
case-sensitive. However, the recognition of the words 'if', 'and' and
'while' is not case-sensitive.

=head2 Constructor

To fill in a template with data, first initialize an B<MCS::CGI::HTML::Template>
object:

    $p = new MCS::CGI::HTML::Template;

The constructor may optionally be initialized with a hash that
associates macro names (the keys) with data (the values). The keys in
the hash do not include the double underscore delimiters:

    $p = new MCS::CGI::HTML::Template HTML_VERSION => 'HTML 3.2 Final',
                            LANGUAGE     => 'EN';

This will cause the strings C<__HTML_VERSION__> and C<__LANGUAGE__> in
the template to be replaced with the strings 'HTML 3.2 Final' and
'EN', respectively.

=head2 Assigning Macro Values

The method B<var> can be used to associate more macro names with
values by passing in a hash:

    $p->var(DATE => scalar(localtime),
            USER => $ENV{REMOTE_USER} || $>);

If B<var> is passed only one argument, then it will return the value
that has been associated with it:

    $user = $p->('USER');

=head2 Conditionals

B<var> is also used to implement conditionals and loops, by making
assignments to the identifiers that appear in an HTML template. For a
conditional, assign a reference to a subroutine to the identifier
associated with C<if> in the template. The subroutine returns a true
value if the conditional text in the HTML template should be
displayed.

For example, the following code will cause a portion of the HTML
template to be generated only if today is a weekday:

  $p->var(WEEKDAY => sub { return (localtime)[6] % 6 != 0 });

  <!-- if WEEKDAY -->
  Today is a weekday!<BR>
  Cheer up anyway!
  <!-- end if -->

=head2 Loops

There are two ways to implement loops. The first is to assign the
identifier a reference to an array of anonymous hashes, where each
hash contains macro assignments that are valid for each iteration of
the loop. In other words, the text defined for the loop in the
template is generated for each element in the array, and each element
contains a 'local' assignment for macros in the loop.

For example, the environment variables can be dumped using this
technique. The data can be prepared as follows:

  @env = ();
  for (keys %ENV) {
    push @env, { VAR => $_, VAL => $ENV{$_} };
  }
  $p->var(ENV => \@env);

The array assigned to C<ENV> now has values for the macros C<__VAR__>
and C<__VAL__> for each iteration of the loop. This could be used with
the following HTML snippet; the text between '<!-- while ENV -->' and
'<!-- end while -->' is repeated for each element of the array.

  <TABLE>
  <!-- while ENV -->
  <TR>
    <TH>__VAR__</TH>
    <TD>__VAL__</TD>
  </TR>
  <!-- end while -->
  </TABLE>

The second way to implement loops is to assign the identifier a
reference to a subroutine. This subroutine returns a true value if the
loop should be continued, and a false value if it should be
terminated. The subroutine can also do anything else you want it to
do, presumably including assignments of data to macros using the
B<var> method.

To dump the environment variables using this technique, we define a
subroutine that uses C<each> to iterate through the hash
(s. L<perlfunc/"each">). The subroutine returns true if there are more
variables to show, false otherwise. On each iteration, it sets the
values of the macros C<VAR> and C<VAL> to the name and value of the
current environment variable. The same HTML code shown above can be
used for this implementation.

  sub env {
      my ($var, $val) = each %ENV;
      if ($var) {
          $p->var(VAR => $var, VAL => $val);
  	  return 1;
      }
      else { return 0; }
  }
  $p->var(ENV => \&env); 

Depending on how you implement it, the approach with a code reference
is likely to be more efficient. To construct an array, you will
probably have to iterate through your data once, and B<MCS::CGI::HTML::Template>
will have to iterate through the array again when the template is
parsed. You will also need to use enough memory to store all of your
data in the array. With a code reference, you can iterate through your
data just once, only using the memory that's needed in each iteration.

=head2 HTML Generation

B<MCS::CGI::HTML::Template> is a subclass of B<HTML::Filter>, which in turn is a
subclass of B<HTML::Parser> (s. L<HTML::Filter>,
L<HTML::Parser>). Thus you can use the B<parse> method with a string
as an argument to generate formatted HTML, and the B<parse_file>
method to format an entire file. The argument to B<parse_file> may be
either a filename or an open filehandle.

    $p->parse('The date and time is <__MARKUP__>__DATE__</__MARKUP__>');

    $p->parse_file('example.tmpl');

    open TEMPLATE, 'example.tmpl'
      or die "Cannot open template: $!\n";
    $p->parse_file(\*TEMPLATE);

The formatted output is sent to standard output.

=pod

=head1 PREREQUISITES

This module is a subclass of B<HTML::Filter>, which must be installed
first. B<HTML::Filter> is currently distributed with the B<LWP> bundle
at CPAN.

=head1 VERSION

$Revision: 1.3 $ $Date: 2007/06/21 13:28:19 $

=head1 AUTHOR

    Geoff Simmons <gs@cityline.net>
    MCS Moorbek Computer System GmbH

=head1 SEE ALSO

L<perl(1)>, L<HTML::Filter(3)>

=cut

##
## Public methods
##

# The constructor initializes an object with HTML::Filter as a
# subclass.
# _ctrl is a stack of identifiers for control structures (IF or WHILE)
# presently being accumulated.
# _line contains the text for a conditional or loop that is presently
# being accumulated.
# _var is a hash containg macro assignments.

sub new {
    my ($class, %params) = @_;

    my $self = new MCS::CGI::HTML::Filter;

    $self->{_ctrl} = [];
    $self->{_line} = '';

    $self->{_var}   = {};
    while (my ($key, $val) = each %params) {
      $self->{_var}{$key} = $val;
    }
    return bless $self, $class;
  }

# If var() gets just one argument, return its value. Otherwise store
# the macro assignments, and return the number of assignments made.

sub var {
  my $self = shift;

  return $self->{_var}{$_[0]} if @_ == 1;

  my %params = @_;
  while (my ($var, $val) = each %params) {
    $self->{_var}{$var} = $val;
  }
  return @_ >> 1;
}

##
## Private methods
##

# The generic subroutine for declarations, start and end tags, and the
# text between tags. If we are presently reading the template for a
# conditional or loop, then add the current text to the template, and
# output nothing.  Otherwise, substitute any macros found in the
# current text, and then hand it off to the superclass.

sub generic {
  my $self     = shift;
  my $origtext = pop;

  if (@{$self->{_ctrl}}) {
    $self->{_line} .= $origtext;
    return @_, '';
  }
  $origtext =~ s/(__(\w+?)__)/defined $self->{_var}{$2}
                                    ? $self->{_var}{$2} : $1/gex;
  return @_, $origtext;
}

sub text        { return $_[0]->SUPER::text(        generic @_ ) }
sub declaration { return $_[0]->SUPER::declaration( generic @_ ) }
sub start       { return $_[0]->SUPER::start(       generic @_ ) }
sub end         { return $_[0]->SUPER::end(         generic @_ ) }

# The comment method does all the real work.

sub comment {
  my ($self, $comment) = @_;
  my $class = ref($self);
  my ($command, $name);

  # If the comment does not consist of a command ('if', 'while' or
  # 'end') followed by an identifier, it's just an ordinary comment
  # and gets the generic treatment.
  if ($comment =~ /^\s*(IF|WHILE|END)\s+(\w+)\s*$/io) {
    ($command, $name) = (uc $1, $2);
  }
  else {
    $comment =~ s/(__(\w+?)__)/defined $self->{_var}{$2}
                                     ? $self->{_var}{$2} : $1/gex;
    return $self->SUPER::comment($comment);
  }

  # For IF or WHILE, push the identifier onto the private stack _ctrl
  # and start storing the text in the private member _line. If we're
  # already in the midst of accumulating text for a control structure,
  # then just keep on accumulating.
  if ($command eq 'IF' or $command eq 'WHILE') {
    if (@{$self->{_ctrl}}) {
      $self->{_line} .= "<!--$comment-->";
    }
    else {
      $self->{_line} = '';
    }
    push @{$self->{_ctrl}}, substr($command,0,1).$name;
    return '';
  }
  else {
    # $command eq 'END'
    $name = uc $name;

    if ($name ne 'IF' and $name ne 'WHILE') {
      # END without IF or WHILE is just an ordinary comment
      if (@{$self->{_ctrl}}) {
	$self->{_line} .= "<!--$comment-->";
	return '';
      }
      $comment =~ s/(__(\w+?)__)/defined $self->{_var}{$2}
	                               ? $self->{_var}{$2} : $1/gex;
      return $self->SUPER::comment($comment);
    }

    # The _ctrl stack enforces balance of IF's and WHILE's. If the top
    # of the stack doesn't match the current END, something's off.
    unless (@{$self->{_ctrl}}) {
      warn "$class (called by: ".(caller(1))[1].") syntax error: 'END $name' found without start\n";
      return $self->SUPER::comment($comment);
    }
    unless ($self->{_ctrl}[-1] =~ /^(.)(.+)$/ and $1 eq substr($name, 0, 1)) {
      warn "$class (called by: ".(caller(1))[1].") syntax error: BEGIN/END $name mismatch\n";
      return $self->SUPER::comment($comment);
    }
    my $struct = $2;
    pop @{$self->{_ctrl}};
    if (@{$self->{_ctrl}}) {
      $self->{_line} .= "<!--$comment-->";
      return '';
    }

    if ($name eq 'WHILE') {
      my $ref  = ref($self->{_var}{$struct});

      # For WHILE with an array reference, fill in the loop line for
      # each hashref in the array. If assignments in the 'local'
      # hashref clash with 'global' assignments, then the local
      # assignment is used.
      # We declare a new parser object to parse the text of the loop
      # so that nesting works.
      if ($ref eq 'ARRAY') {
	for (@{$self->{_var}{$struct}}) {
	  my $p = new $class;
	  $p->{_var} = { %{$self->{_var}}, %{$_} };
	  $p->parse($self->{_line});
	  $p->eof();
	}
	$self->{_line} = '';
	return '';
      }
      elsif ($ref eq 'CODE') {
        # For WHILE with a code reference, fill in the loop line every
        # time the coderef returns true.
	while (&{$self->{_var}{$struct}}) {
	  my $p = new $class;
	  $p->{_var} = $self->{_var};
	  $p->parse($self->{_line});
	  $p->eof();
	}
	$self->{_line} = '';
	return '';
      }
      else {
	warn
	  "$class (called by: ".(caller(1))[1].") error: ",
	  "No code or array reference provided for WHILE $struct\n";
	my $line = $self->{_line};
	$line =~ s/(__(\w+?)__)/defined $self->{_var}{$2}
                                      ? $self->{_var}{$2} : $1/gex;
	$self->SUPER::output($line);
	$self->{_line} = '';
	return '';
      }
    }
    elsif ($name eq 'IF') {
      # For an IF, we just evaluate the coderef and generate the
      # conditional text if it returns true.
      unless (ref($self->{_var}{$struct}) eq 'CODE') {
	warn "$class (".(caller(1))[1].") error: No code reference provided for IF $struct\n";
	my $line = $self->{_line};
	$line =~ s/(__(\w+?)__)/defined $self->{_var}{$2}
	                              ? $self->{_var}{$2} : $1/gex;
	$self->SUPER::output($line);
	$self->{_line} = '';
	return '';
      }
      if (&{$self->{_var}{$struct}}) {
	my $p = new $class;
	$p->{_var} = $self->{_var};
	$p->parse($self->{_line});
	$p->eof();
      }
      $self->{_line} = '';
      return ''
    }
  }
}

1;

__END__
